/**
 * Ein Knoten eines allgemeinen AVL-Baums.
 * 
 * @author Bernd Gramlich 
 * @version 2012-12-21
 */
public class Knoten extends Baumelement
{
    /*
     * Inhalt des Knotens ist ein Eintrag.
     */
    private Eintrag ich;

    /*
     * Nachfolger des Knotens sind zwei Baumelemente,
     * also entweder wieder Knoten oder Abschlsse
     * des Baums.
     */
    private Baumelement linkerNachfolger;
    private Baumelement rechterNachfolger;
    
    /*
     * Bei einem AVL-Baum drfen sich nach jedem Einfge-
     * und Lschschritt an jedem Knoten die linke und die
     * rechte Hhe um hchstens 1 unterscheiden. Wir
     * speichern diese Hhen als Attribute, um sie nicht
     * jedesmal neu rekursiv berechnen zu mssen.
     */
    private int linkeHoehe;
    private int rechteHoehe;

    /**
     * Konstruktor.
     * Erzeuge einen neuen Knoten mit Inhalt "meinInhalt" und
     * Nachfolgern "meinLinkerNachfolger" und "meinRechterNachfolger"
     * an der "baumposition".
     */
    public Knoten (Baumelement meinLinkerNachfolger, Baumelement meinRechterNachfolger,
                   Eintrag meinInhalt, String baumposition)
    {
        linkerNachfolger  = meinLinkerNachfolger;
        rechterNachfolger = meinRechterNachfolger;
        ich               = meinInhalt;

        ich.zeichneAnBaumposition (baumposition);
        
        berechneLinkeHoehe ();
        berechneRechteHoehe ();
    }

    /**
     * Nenne den linken Nachfolger des Knotens.
     */
    public Baumelement gibLinkenNachfolger ()
    {
        return linkerNachfolger;
    }

    /**
     * Nenne den rechten Nachfolger des Knotens.
     */
    public Baumelement gibRechtenNachfolger ()
    {
        return rechterNachfolger;
    }

    /**
     * Nenne den Inhalt des Knotens.
     */
    public Eintrag gibInhalt ()
    {
        return ich;
    }

    /**
     * Nenne die Hhe meines Teilbaums.
     * Meine Hhe ist um 1 grer als die grere der
     * beiden Hhen in meinem linken und meinem
     * rechten Teilbaum.
     */
    public int gibHoehe ()
    {
        return 1 + Math.max (linkeHoehe, rechteHoehe);
    }

    /**
     * Nenne die AVL-Balance meines Teilbaums.
     */
    public int gibBalance ()
    {
        return linkeHoehe - rechteHoehe;
    }

    /**
     * Verschaffe dem Knoten einen neuen linken Nachfolger.
     */
    public void setzeLinkenNachfolger (Baumelement neuerLinkerNachfolger)
    {
        linkerNachfolger = neuerLinkerNachfolger;
        /*
         * Man knnte hier auch gleich die linke Hhe
         * neu berechnen.
         */
    }

    /**
     * Verschaffe dem Knoten einen neuen rechten Nachfolger.
     */
    public void setzeRechtenNachfolger (Baumelement neuerRechterNachfolger)
    {
        rechterNachfolger = neuerRechterNachfolger;
        /*
         * Man knnte hier auch gleich die linke Hhe
         * neu berechnen.
         */
    }
    
    /**
     * Berechne meine linke Hoehe.
     */
    public void berechneLinkeHoehe ()
    {
        linkeHoehe = linkerNachfolger.gibHoehe ();
    }
    
    /**
     * Berechne meine rechte Hoehe.
     */
    public void berechneRechteHoehe ()
    {
        rechteHoehe = rechterNachfolger.gibHoehe ();
    }

    /**
     * Fge einen neuen Eintrag aufsteigend sortiert in den Baum ein.
     */
    public Baumelement fuegePassendEin (Eintrag neuerEintrag, String baumposition)
    {
        if (ich.istKleinerAls (neuerEintrag))
        {
            /*
             * Wenn ich kleiner bin als der neue Eintrag, dann muss
             * der neue Eintrag zu meiner Rechten eingefgt werden. Diese
             * Aufgabe berlasse ich meinem rechten Nachfolger.
             */
            rechterNachfolger = rechterNachfolger.fuegePassendEin (neuerEintrag, baumposition + "r");
            berechneRechteHoehe ();
        }
        else
        {
            /*
             * Wenn der neue Eintrag hchstens so gro ist wie ich,
             * dann muss der neue Eintrag zu meiner Linken
             * eingefgt werden. Diese Aufgabe berlasse ich meinem
             * linken Nachfolger.
             */
            linkerNachfolger = linkerNachfolger.fuegePassendEin (neuerEintrag, baumposition + "l");
            berechneLinkeHoehe ();
        }
            
        /*
         * Ich balanciere meinen Teilbaum ntigenfalls aus.
         */ 
        return balanciere (baumposition);
    }

    /**
     * Drucke den Inhalt des Binrbaums ab mir selbst in
     * aufsteigender Reihenfolge auf die Konsole.
     */
    public void druckeInOrder ()
    {
        linkerNachfolger.druckeInOrder ();
        ich.druckeAufKonsole ();
        rechterNachfolger.druckeInOrder ();
    }

    /**
     * Drucke den Inhalt des Binrbaums ab mir selbst in
     * Prordnung auf die Konsole.
     */
    public void druckePreOrder ()
    {
        ich.druckeAufKonsole ();
        linkerNachfolger.druckePreOrder ();
        rechterNachfolger.druckePreOrder ();
    }

    /**
     * Drucke den Inhalt des Binrbaums ab mir selbst in
     * Postordnung auf die Konsole.
     */
    public void druckePostOrder ()
    {
        linkerNachfolger.druckePostOrder ();
        rechterNachfolger.druckePostOrder ();
        ich.druckeAufKonsole ();
    }

    /**
     * Zhle, wie viele Knoten ab mir selbst die "eigenschaft" haben.
     */
    public int zaehle (Eigenschaft eigenschaft)
    {
        if (eigenschaft.hatEigenschaft (ich))
        {
            return 1 + linkerNachfolger.zaehle (eigenschaft)
                     + rechterNachfolger.zaehle (eigenschaft);
        } 
        else
        {
            return linkerNachfolger.zaehle (eigenschaft)
                   + rechterNachfolger.zaehle (eigenschaft);
        }
    }

    /**
     * Fhre fr alle Knoten ab mir selbst, die die Eigenschaft
     * "eigenschaft" haben, den Befehl "befehl" mit den Argumenten
     * "zeichenkette" aus.
     */
    public void fuehreAusFuer (Eigenschaft eigenschaft, Befehl befehl, String[] zeichenkette)
    {
        if (eigenschaft.hatEigenschaft (ich))
        {
            befehl.fuehreAus (ich, zeichenkette);
        }

        linkerNachfolger.fuehreAusFuer (eigenschaft, befehl, zeichenkette);
        rechterNachfolger.fuehreAusFuer (eigenschaft, befehl, zeichenkette);
    }

    /**
     * Suche einen Eintrag ab mir im Baum (in order), der zum Suchwort passt.
     */
    public Eintrag suche (String suchwort)
    {
        if (ich.vergleicheMit (suchwort) > 0)
        {
            return linkerNachfolger.suche (suchwort);
        }
        else if (ich.vergleicheMit (suchwort) == 0)
        {
            return ich;
        }
        else
        {
            return rechterNachfolger.suche (suchwort);
        }
    }

    /**
     * Lsche den ersten Knoten ab mir im Baum, der zum Suchwort passt.
     */
    public Baumelement loesche (String suchwort, String baumposition)
    {
        if (ich.vergleicheMit (suchwort) > 0)
        {
            /*
             * Wenn ich grer bin als das Suchwort, dann
             * lsche in meinem linken Teilbaum.
             */
            linkerNachfolger = linkerNachfolger.loesche (suchwort, baumposition + "l");
            berechneLinkeHoehe ();
        }
        else if (ich.vergleicheMit (suchwort) == 0)
        {
            /*
             * Wenn ich gleich dem Suchwort bin, dann muss
             * ich mich selbst lschen. Dazu mache ich mich
             * zunchst einmal unsichtbar.
             */
            ich.macheUnsichtbar ();

            if (linkerNachfolger instanceof Abschluss)
            {
                /*
                 * Wenn ich keinen linken Teilbaum habe, dann
                 * lsche ich mich, indem ich mich durch meinen
                 * rechten Nachfolger ersetze. Der Teilbaum
                 * meines rechten Nachfolgers muss dabei an meiner
                 * Baumposition neu gezeichnet werden.
                 */
                rechterNachfolger.zeichneNeu (baumposition);

                return rechterNachfolger;
            }
            else if (rechterNachfolger instanceof Abschluss)
            {
                /*
                 * Wenn ich keinen rechten Teilbaum habe, dann
                 * lsche ich mich, indem ich mich durch meinen
                 * linken Nachfolger ersetze. Der Teilbaum
                 * meines linken Nachfolgers muss dabei an meiner
                 * Baumposition neu gezeichnet werden.
                 */
                linkerNachfolger.zeichneNeu (baumposition);

                return linkerNachfolger;
            }
            else
            {
                /*
                 * Wenn ich einen linken und einen rechten Teilbaum
                 * habe, dann ersetze ich mich durch das Maximum in
                 * meinem linken Teilbaum.
                 */
                ich = linkerNachfolger.gibMaximum ();
                ich.zeichneAnBaumposition (baumposition);
                
                /*
                 * Nun muss ich in meinem linken Nachfolger das Maximum
                 * lschen, weil es ja jetzt in meinem Knoten steht.
                 */
                linkerNachfolger = linkerNachfolger.loescheMaximum (baumposition + "l");
                berechneLinkeHoehe ();
                
                ich.zeichneAnBaumposition (baumposition);
            }
        }
        else
        {
            /*
             * Wenn ich kleiner bin als das Suchwort, dann
             * lsche in meinem rechten Teilbaum.
             */
            rechterNachfolger = rechterNachfolger.loesche (suchwort, baumposition + "r");
            berechneRechteHoehe ();
        }
            
        /*
         * Ich balanciere meinen Teilbaum ntigenfalls aus.
         */ 
        return balanciere (baumposition);
    }

    /**
     * Nenne den grten Eintrag ab mir im Baum.
     */
    public Eintrag gibMaximum ()
    {
        if (rechterNachfolger instanceof Abschluss)
        {
            /*
             * Wenn mein rechter Nachfolger ein Abschluss ist,
             * dann bin ich der maximale Knoten in meinem Teilbaum.
             */
            return ich;
        }
        else
        {
            /*
             *Andernfalls befindet sich der maximale Knoten im
             *Teilbaum meines rechten Nachfolgers.
             */
            return rechterNachfolger.gibMaximum ();
        }
    }

    /**
     * Lsche den grten Knoten ab mir im Baum.
     */
    public Baumelement loescheMaximum (String baumposition)
    {
        if (rechterNachfolger instanceof Abschluss)
        {
            /*
             * Wenn mein rechter Nachfolger ein Abschluss ist,
             * dann bin ich der maximale Knoten in meinem Teilbaum.
             * Ich lsche mich, indem ich meinen linken Nachfolger
             * an meine Stelle aufrcken lasse.
             */
            ich.macheUnsichtbar ();
            linkerNachfolger.zeichneNeu (baumposition);
            
            return linkerNachfolger;
        }
        else
        {
            /*
             * Andernfalls befindet sich der zu lschende maximale
             * Knoten im Teilbaum meines rechten Nachfolgers.
             */
            rechterNachfolger = rechterNachfolger.loescheMaximum (baumposition + "r");
            berechneRechteHoehe ();
            
            /*
             * Ich balanciere meinen Teilbaum ntigenfalls aus.
             */ 
            return balanciere (baumposition);
        }
    }

    /**
     * Zeichne den Baum ab mir neu.
     */
    public void zeichneNeu (String baumposition)
    {
        ich.macheUnsichtbar ();
        ich.zeichneAnBaumposition (baumposition);

        linkerNachfolger.zeichneNeu (baumposition + "l");
        rechterNachfolger.zeichneNeu (baumposition + "r");
    }
    
    /**
     * Rotiere nach links.
     */
    public Baumelement rotiereNachLinks (String baumposition)
    {
        /*
         * Mein rechter Nachfolger soll neue Wurzel in meinem
         * Teilbaum werden.
         */
        Knoten neuerHauptknoten = (Knoten) rechterNachfolger;
        
        /*
         * Ich bernehme den linken Nachfolger meines rechten
         * Nachfolgers als meinen neuen rechten Nachfolger.
         */
        rechterNachfolger = neuerHauptknoten.gibLinkenNachfolger ();
        
        /*
         * Ich werde linker Nachfolger meines ehemaligen rechten
         * Nachfolgers.
         */
        neuerHauptknoten.setzeLinkenNachfolger (this);
        
        /*
         * Ich passe die Hhen an und zeichne meinen Teilbaum neu.
         */
        berechneRechteHoehe ();
        neuerHauptknoten.berechneLinkeHoehe ();
        
        neuerHauptknoten.zeichneNeu (baumposition);
        
        /*
         * Ich melde, dass mein ehemaliger rechter Nachfolger neue
         * Wurzel meines Teilbaums ist.
         */
        return neuerHauptknoten;
    }
    
    /**
     * Rotiere nach rechts.
     */
    public Baumelement rotiereNachRechts (String baumposition)
    {
        /*
         * Mein linker Nachfolger soll neue Wurzel in meinem
         * Teilbaum werden.
         */
        Knoten neuerHauptknoten = (Knoten) linkerNachfolger;
        
        /*
         * Ich bernehme den rechten Nachfolger meines linken
         * Nachfolgers als meinen neuen linkenen Nachfolger.
         */
        linkerNachfolger = neuerHauptknoten.gibRechtenNachfolger ();
        
        /*
         * Ich werde rechter Nachfolger meines ehemaligen linken
         * Nachfolgers.
         */
        neuerHauptknoten.setzeRechtenNachfolger (this);
                 
        /*
         * Ich passe die Hhen an und zeichne meinen Teilbaum neu.
         */
        berechneLinkeHoehe ();
        neuerHauptknoten.berechneRechteHoehe ();
        
        neuerHauptknoten.zeichneNeu (baumposition);
        
        /*
         * Ich melde, dass mein ehemaliger linker Nachfolger neue
         * Wurzel meines Teilbaums ist.
         */
        return neuerHauptknoten;
    }
    
    /**
     * Balanciere meinen Teilbaum aus.
     * Diese Methode muss beim Einfgen und beim Lschen aufgerufen
     * werden, um sicherzustellen, dass die AVL-Bedingung weiterhin
     * erfllt ist.
     */
    private Baumelement balanciere (String baumposition)
    {
        /*
         * Wenn meine Balance -2 oder 2 ist, muss gem den AVL-Regeln
         * rotiert werden.
         */
        switch (gibBalance ())
        {
            case -2:
            /*
             * Wenn meine AVL-Balance -2 ist, muss ich nach links
             * rotieren.
             * 
             * Zuvor muss ich allerdings verhindern, dass mein
             * rechter Nachfolger ein anderes Vorzeichen in seiner
             * AVL-Balance hat als ich. Notfalls muss ich ihn also
             * zuvor nach rechts rotieren.
             */
            if (rechterNachfolger.gibBalance () == 1)
            {
                rechterNachfolger = rechterNachfolger.rotiereNachRechts (baumposition + "r");
                berechneRechteHoehe ();
            }
                
            return rotiereNachLinks (baumposition);
            
            case 2:
            /*
             * Wenn meine AVL-Balance 2 ist, muss ich nach rechts
             * rotieren.
             * 
             * Zuvor muss ich allerdings verhindern, dass mein
             * linker Nachfolger ein anderes Vorzeichen in seiner
             * AVL-Balance hat als ich. Notfalls muss ich ihn also
             * zuvor nach rechts rotieren.
             */
            if (linkerNachfolger.gibBalance () == -1)
            {
                linkerNachfolger = linkerNachfolger.rotiereNachLinks (baumposition + "l");
                berechneLinkeHoehe ();
            }
                
            return rotiereNachRechts (baumposition);
            
            default:
            /*
             * Wenn meine AVL-Balance -1; 0 oder 1 ist,
             * muss ich nicht rotieren.
             */
            return this;
        }
    }
}
